import base64
import datetime
import decimal

import pytest
import pytz
import requests

import festung.dbapi
from festung import exceptions
from festung.types import Type


@pytest.fixture
def connection_session():
    with requests.Session() as session:
        yield session


@pytest.fixture(params=['managed', 'external'])
def connection_kwargs(request, connection_session):
    if request.param == 'managed':
        return {}

    elif request.param == 'external':
        return dict(session=connection_session)

    else:
        assert False, "Not all parameters are supported"


@pytest.fixture
def connection(database_url, connection_kwargs):
    conn = festung.dbapi.Connection(database_url, **connection_kwargs)
    try:
        yield conn
    finally:
        conn.close()


@pytest.fixture
def cursor(connection):
    cur = connection.cursor()
    try:
        yield cur
    finally:
        cur.close()


class ResponseFixture(object):
    @pytest.fixture(autouse=True)
    def prepare_responses(self, festung):
        if not self.responses:
            return
        for response in self.responses:
            festung.add_response(response)

    @property
    def data(self):
        return [response['data'] for response in self.responses]

    @property
    def first(self):
        try:
            return self.data[0]
        except IndexError:
            pass


class ErrorResponseFixture(ResponseFixture):

    # XXX(moritz): This is to circumvent a misbehavior in pytest
    data = None
    first = None


class TestEmptyResponse(ResponseFixture):
    responses = [{'data': [], 'headers': [], 'last_row_id': 0, 'rows_changed': 0}]

    def test_execute(self, festung, cursor):
        query = 'SELECT ?, ?, ?'
        params = ['string', 1, None]
        cursor.execute(query, params)

        [data] = festung.json_queries
        assert data['sql'] == query
        assert data['params'] == params
        assert cursor.lastrowid == self.responses[0]['last_row_id']
        assert cursor.rowcount == self.responses[0]['rows_changed']

    @pytest.mark.parametrize('param,expected_serialialized', [
        (datetime.datetime(1970, 1, 1, 0, 0, tzinfo=pytz.UTC), '1970-01-01T00:00:00+00:00'),
        (datetime.date(1970, 1, 1), '1970-01-01'),
        (datetime.time(0, 0, 0), '00:00:00'),
        (decimal.Decimal('1234.5678'), '1234.5678'),
    ])
    def test_binding_cast(self, festung, cursor, param, expected_serialialized):
        cursor.execute('SELECT 1', [param])
        [data] = festung.json_queries
        [serialized] = data['params']

        assert serialized == expected_serialialized

    def test_password_is_sent(self, cursor, password, festung):
        cursor.execute('SELECT * FROM foo')
        [query] = festung.queries
        assert base64.b64decode(query.headers['Authorization']) == password


class TestMultipleResponses(object):

    def test_executemany(self, festung, cursor):
        for _ in range(3):
            festung.add_response({'data': [], 'headers': [], 'last_row_id': 0, 'rows_changed': 3})

        query = 'UPDATE TABLE foo SET foo=?, bar=?, baz=?'
        params_list = [
            ['abc', 1, None],
            ['def', 2, 1.2],
            ['ghi', 3, None],
        ]
        cursor.executemany(query, params_list)
        for params, data in zip(params_list, festung.json_queries):
            assert data['sql'] == query
            assert data['params'] == params

    def test_executemany_with_result(self, festung, cursor):
        festung.add_response({
            'data': [[1]],
            'headers': [{'name': 'foo', 'type': 'int'}],
            'last_row_id': 0,
            'rows_changed': 0,
        })
        query = 'SELECT ?'
        params_list = [[1]]
        with pytest.raises(exceptions.ProgrammingError):
            cursor.executemany(query, params_list)

    @pytest.mark.parametrize('given,expected', [
        ('int', Type.NUMBER),
        ('blob', Type.BINARY),
        ('numeric(19, 2)', Type.NUMBER),
        ('dynamic', Type.UNKNOWN),
    ])
    def test_types(self, festung, cursor, given, expected):
        festung.add_response({
            'data': [],
            'headers': [{'name': 'foo', 'type': given}],
            'last_row_id': 0,
            'rows_changed': 0,
        })
        query = 'SELECT foo'
        cursor.execute(query)
        assert cursor.description[0].type_code == expected


class TestDummyResponseWithExecutedCursor(ResponseFixture):
    responses = [
        {
            'data': [['a', 1], ['b', 2]],
            'headers': [{'name': 'bar', 'type': 'string'}, {'name': 'baz', 'type': 'int'}],
            'last_row_id': 0,
            'rows_changed': 0,
        },
    ]

    @pytest.fixture(autouse=True)
    def run_query(self, cursor):
        cursor.execute("SELECT * FROM foo")

    def test_fetchone(self, cursor):
        assert cursor.fetchone() == tuple(self.first[0])
        assert cursor.fetchone() == tuple(self.first[1])
        assert cursor.fetchone() is None

    def test_fetchall_same_size_than_response(self, cursor):
        cursor.arraysize = len(self.first)
        assert cursor.fetchmany() == [tuple(row) for row in self.first]

    def test_fetchall_smaller_size_than_response(self, cursor):
        cursor.arraysize = 1
        assert cursor.fetchmany() == [tuple(self.first[0])]

    def test_fetchall_bigger_size_than_response(self, cursor):
        cursor.arraysize = len(self.first) * 10
        assert cursor.fetchmany() == [tuple(row) for row in self.first]


class TestErroneousResponse(ErrorResponseFixture):
    responses = [
        {
            'errors': ['Missing Header "Authorization"'],
            'message': "Invalid use of endpoint.",
        },
    ]

    def test_execute(self, festung, cursor):
        query = 'SELECT 1;'
        with pytest.raises(exceptions.DataError) as excinfo:
            cursor.execute(query)
            e = excinfo.value
            [response] = self.responses
            assert e.message == response['message']
            assert e.errors == response['errors']



import base64
from collections import namedtuple
import urlparse

import requests

from festung._private import cast
from festung._private import no_password_url
from festung._private import to_http_url
# Exceptions have to be on the DBAPI module
from festung.exceptions import DatabaseError      # NOQA
from festung.exceptions import DataError          # NOQA
from festung.exceptions import Error              # NOQA
from festung.exceptions import IntegrityError     # NOQA
from festung.exceptions import InterfaceError     # NOQA
from festung.exceptions import InternalError      # NOQA
from festung.exceptions import NotSupportedError  # NOQA
from festung.exceptions import OperationalError   # NOQA
from festung.exceptions import ProgrammingError   # NOQA
from festung.exceptions import Warning            # NOQA
import festung.pool
import festung.types
# Types have to be on the DBAPI module
from festung.types import BINARY                  # NOQA
from festung.types import Binary                  # NOQA
from festung.types import Date                    # NOQA
from festung.types import DateFromTicks           # NOQA
from festung.types import DATETIME                # NOQA
from festung.types import NUMBER                  # NOQA
from festung.types import ROWID                   # NOQA
from festung.types import STRING                  # NOQA
from festung.types import Time                    # NOQA
from festung.types import TimeFromTicks           # NOQA
from festung.types import Timestamp               # NOQA


__all__ = ['connect', 'apilevel', 'paramstyle', 'threadsafety', 'Connection', 'Cursor']


apilevel = '2.0'
threadsafety = 3  # Threads may share the module, connections and cursors
paramstyle = 'qmark'

SCHEME = 'festung'


class Connection(object):
    def __init__(self, url, session=None):
        parsed_url = urlparse.urlparse(url)
        if parsed_url.scheme != SCHEME:
            raise ValueError("We only support festung:// connections.")
        self.url = url
        self.pool = festung.pool.get_pool(session)

    def close(self):
        self.pool.close()

    def commit(self):
        # TODO(Antoine): Implement
        pass

    def rollback(self):
        # TODO(Antoine): Implement
        pass

    def cursor(self):
        return Cursor(self)

    @property
    def _http_url(self):
        return to_http_url(self.url)

    @property
    def _password(self):
        parsed_url = urlparse.urlsplit(self.url)
        return parsed_url.password or ''

    def _request(self, method, **kwargs):
        req = requests.Request(method, self._http_url, **kwargs)
        req.headers['Authorization'] = base64.b64encode(self._password)
        resp = self.pool.request(req)
        return resp.json()

    def __repr__(self):
        return "<Connection({})>".format(no_password_url(self.url))


connect = Connection


CursorDescription = namedtuple(
    'CursorDescription', 'name,type_code,display_size,internal_size,precisison,scale,null_ok')


NO_EXECUTE_ROWCOUNT = -1  # <https://www.python.org/dev/peps/pep-0249/#rowcount>
NO_EXECUTE_DESCRIPTION = None  # <https://www.python.org/dev/peps/pep-0249/#description>
NO_MORE_ROW = None  # <https://www.python.org/dev/peps/pep-0249/#fetchone>
NO_EXECUTE_ITER = object()


def _generate_description(headers):
    names = [h['name'] for h in headers]
    # FIXME(Antoine): This is not row you're supposed to infer SQL types
    type_codes = [festung.types.Type.from_header(h['type']) for h in headers]

    # TODO(Antoine): Read more description attributes
    all_none = [None] * len(headers)
    display_sizes = all_none
    internal_sizes = all_none
    precisions = all_none
    scales = all_none
    null_ok = all_none

    all_args = zip(names, type_codes, display_sizes, internal_sizes, precisions, scales, null_ok)
    return [CursorDescription(*args) for args in all_args]


# FIXME(Antoine): Support contextmanager interface (for .close())
class Cursor(object):
    def __init__(self, connection):
        self.connection = connection
        self._iter = NO_EXECUTE_ITER
        self._description = NO_EXECUTE_DESCRIPTION
        self._rowcount = NO_EXECUTE_ROWCOUNT
        self._arraysize = 1
        self.lastrowid = 0

    @property
    def description(self):
        return self._description

    @property
    def rowcount(self):
        return self._rowcount

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value):
        if not isinstance(value, (int, long)):
                raise ValueError("arraysize should be an integer")
        self._arraysize = value

    def callproc(self, procname, parameters=None):
        parameters = parameters or []
        raise NotImplementedError

    def execute(self, operation, parameters=None):
        parameters = parameters or []
        data = dict(sql=operation, params=[cast(p) for p in parameters])
        res = self._request('POST', json=data)
        # For debugging purposes
        self._data = res
        # FIXME(moritz): Also handle if res contains `type` (i.e. operational errors)
        if 'errors' in res:
            raise DataError(res['message'], errors=res['errors'])
        self._iter = iter(res['data'])
        self.lastrowid = res['last_row_id']
        # FIXME(Antoine): I'm not sure that's the way to implement rowcount...
        self._rowcount = res['rows_changed']

        self._description = _generate_description(res['headers'])

    def executemany(self, operation, parameters_sequence):
        for parameters in parameters_sequence:
            self.execute(operation, parameters)
            if self.fetchone() is not NO_MORE_ROW:
                raise ProgrammingError("The statement shall not produce a result.")

    def fetchone(self):
        if self._iter is NO_EXECUTE_ITER:
            raise ProgrammingError("No statement was executed on this cursor.")
        try:
            return tuple(next(self._iter))
        except StopIteration:
            return NO_MORE_ROW

    def fetchmany(self, size=None):
        if size is None:
            size = self.arraysize

        acc = []
        for _ in range(size):
            row = self.fetchone()
            if row is NO_MORE_ROW:
                break
            acc.append(row)
        return acc

    def fetchall(self):
        # FIXME(Antoine): This is very similar to fetchmany(), this should be
        #                 refactor in one function.
        acc = []
        while True:
            row = self.fetchone()
            if row is NO_MORE_ROW:
                break
            acc.append(row)
        return acc

    def nextset(self):
        raise NotImplementedError

    def setinputsize(sizes):
        raise NotImplementedError

    def setoutputsize(size, columns=None):
        raise NotImplementedError

    @property
    def rownumber(self):
        raise NotImplementedError

    def close(self):
        pass

    def _request(self, *args, **kwargs):
        return self.connection._request(*args, **kwargs)



import datetime
import re
import time

import enum


Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime
Binary = buffer


def DateFromTicks(ticks):
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    return Timestamp(*time.localtime(ticks)[:6])


class Type(enum.Enum):
    STRING = 'string'
    BINARY = 'binary'
    NUMBER = 'number'
    DATETIME = 'datetime'
    ROWID = 'rowid'
    UNKNOWN = 'unknown'

    @classmethod
    def from_header(cls, header):
        header = _clean_header(header)
        header = header.lower()
        try:
            return TYPE_REMAPPING[header]
        except KeyError:
            return cls(header)


header_cleaner = re.compile(r'^(\w+)\(.*\)$')


def _clean_header(header):
    match = header_cleaner.match(header)
    if match:
        header = match.group(1)
    return header


TYPE_REMAPPING = {
    'bigint': Type.NUMBER,
    'blob': Type.BINARY,
    'boolean': Type.NUMBER,
    'cblob': Type.STRING,
    'character': Type.STRING,
    'datetime': Type.DATETIME,
    'date': Type.DATETIME,
    'decimal': Type.NUMBER,
    'double precisison': Type.NUMBER,
    'double': Type.NUMBER,
    'dynamic': Type.UNKNOWN,
    'float': Type.NUMBER,
    'int2': Type.NUMBER,
    'int8': Type.NUMBER,
    'integer': Type.NUMBER,
    'int': Type.NUMBER,
    'mediumint': Type.NUMBER,
    'native character': Type.STRING,
    'nchar': Type.STRING,
    'numeric': Type.NUMBER,
    'nvarchar': Type.STRING,
    'real': Type.NUMBER,
    'smallint': Type.NUMBER,
    'text': Type.STRING,
    'tinyint': Type.NUMBER,
    'unsigned big int': Type.NUMBER,
    'varchar': Type.STRING,
    'varying character': Type.STRING,
}


STRING = Type.STRING
BINARY = Type.BINARY
NUMBER = Type.NUMBER
DATETIME = Type.DATETIME
ROWID = Type.ROWID
